/*
    =========================================================================
	File:		DependencyGraph.java
	Author:		Warren Pfeffer
	Contact:	slinky@njcc.com
    =========================================================================
    Copyright (C) 1997  Warren Pfeffer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program (see COPYING-2.0.gz); if not, write to:

		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307  USA
    =========================================================================
*/
package WARREN.tools;

import java.io.InputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.FilenameFilter;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.util.Vector;
import java.lang.reflect.Field;
import COM.objectspace.jgl.HashMap;
import COM.objectspace.jgl.HashSet;

//------------------------------------------

public class DependencyGraph extends JDSystem {

	public static class Props extends GetProps {
		public boolean dbug;	// general debugging on or off
		public boolean prof;	// turn profiling on or off
		public boolean nozip;	// show zip contents as the dependency
		public boolean noinner;	// toggle dipslay of inner class targets
		public boolean noescinner;// use $$ for inner classes
		public boolean shwclspth; // display CLASSPATH as comment
		public boolean srcgph;	// display a source dependency graph
		public boolean nowarn;	// supress warnings
		public boolean noclspth; // do not resolve deps from CLASSPATH 
		public boolean nowrap;	
		public String  rules;	// specify make action rules
		public String  pkgmnt;	// path prefix for current .java package
		public String  abspkgmnt; // use absolute path prefix for current .java package
		public String  xpkgs;	// no dependencies on these packages
		public String  allrule; // set target name of the all rule 
		public String  extpkgs;
		public String  prerules;
		public String  postrules;
	}
	private static Props $ = new Props();

	private	StrBuf			all_rule_;
	private int				all_rule_start_len_;
	private	StrBuf			action_rules_;
	private	StrBuf			pre_rules_;
	private	StrBuf			post_rules_;
	private HashSet			xcld_pkgs_;
	private String			file_sep_;
	private String			path_sep_;
	private String			class_path_;
	private String			cwd_;		// current dir
	private SequencedSet	class_dir_set_;
	private HashMap			imports_;	// hash of imported package locations 
	private String			pkg_mnt_;
	private String			pkg_mnt_canon_;
	private int				add_import_called_;
	private boolean			use_abspkgmnt_;
	private Vector			ext_nodes_;
	private HashMap			ext_pkgs_;
	private SrcDG			srcdg_;
	private SequencedSet	srcdg_set_;
	private HashMap			non_pkg_cls_dir_map_;
	private ClsSfxFilter	cls_sfx_filter_;
	private HashMap			tgts_to_src_map_;
	private HashMap			unpfxd_tgt_to_pfxd_tgt_map_;
	private StrBuf			warnings_;

	//--------------------------------------------------------------

	class ClsSfxFilter implements FilenameFilter {

		public final boolean accept(File dir, String name)
		{
			return (name.endsWith(".class"));
		}
	}

	//--------------------------------------------------------------

	class SrcDG {

		String			file_name;
		String			pkg_name;
		String			pkg_pth;
		Pkg				unres_pkg;
		SequencedSet	targets;
		HashSet			var_set;
		SequencedSet	dep_set;
		StrBuf			warnings;

		SrcDG(String file_name)
		{
			this.file_name = file_name;
			dep_set = new SequencedSet();
			pkg_name = "";
			pkg_pth = "";
			unres_pkg = null;
			targets = null;
			var_set = null;
			warnings = new StrBuf("");
		}

		public int hashCode()
			{return file_name.hashCode();}

		public String toString()
			{return file_name;}

		public boolean equals(Object obj)
			{return file_name.equals(obj.toString());}
	}
	//--------------------------------------------------------------

	public void Reset()
	{
		srcdg_ = null;
		add_import_called_ = 0;
	}

	public void WarningsAppend(String warning)
		{warnings_.Append("# WARNING: "+warning+"\n");}

	public void WarningsAppend(SrcDG srcdg, String warning)
		{srcdg.warnings.Append("# WARNING: "+warning+"\n");}

	public void ResolveDeps()
	{
		SrcDG	srcdg = null;
		SrcDG	srcdg_dep = null;
		String	src_name = null;
		String	pkg_name = null;
		Dep		dep = null;
		Tgt		tgt = null;

		for(
			Enumeration s_itr = srcdg_set_.Iterator();
			s_itr.hasMoreElements();
			)
		{
			srcdg = (SrcDG)s_itr.nextElement();

			if($.dbug)
				PrintDbugln("srcdg.file_name="+srcdg.file_name);

			for(
				Enumeration itr = srcdg.dep_set.Iterator();
				itr.hasMoreElements();
			   )
			{
				dep = (Dep)itr.nextElement();

				if(dep.State() != Dep.RESOLVED)
				{
					// remember only .class deps can be unresolved

					String dep_name = dep.NoSfxName(".class");

					if($.dbug)
						PrintDbugln("ResolveDeps: dep_name="+dep_name);

					srcdg_dep = null;
					int di = 0;

					for(;;)
					{
						if(
							((srcdg_dep = FindSrcFromCls(dep_name+".class")) != null)
							||
							((di = dep_name.lastIndexOf('.')) < 1)
						  )
							break;

						dep_name = ReplaceLast(dep_name, ".", "$");
					}

					if(srcdg_dep == null)
					{
						if(di < 1)
						{
							srcdg_dep =
								FindSrcFromCls(FilterClsFromImps(dep_name+".class"));
						}
						else
						{
							// try local package 

							if(srcdg.pkg_name.length() > 0)
							{
								pkg_name = srcdg.pkg_name+".";

								srcdg_dep =
									FindSrcFromCls(pkg_name+dep_name+".class");
							}
						}
					}

					if(srcdg_dep != null)
					{
						// dep is a class

						di = dep_name.lastIndexOf('.');

						if(di > 0)
							dep_name = dep_name.substring(di+1);
						
						dep.Resolve(srcdg_dep.pkg_pth+dep_name+".class");
					}
					else
					{
						String base_cls = dep.GetType().BaseCls();
						Var var = null;
						String var_name = null;

						while((srcdg_dep = FindSrcFromCls(FilterClsFromImps(base_cls+".class"))) != null)
						{
							var = (Var)srcdg_dep.var_set.get(dep.NoSfxName(".class"));
							if(var != null && var.DeclCls().equals(base_cls))
							{
								// ignore dep that is var from base class

								if($.dbug)
									PrintDbugln("marked for ignore ["+dep.Name()+"]");

								dep.SetState(Dep.IGNORE);
								break;
							}
							else
							{
								if($.dbug)
									PrintDbugln("ResolveDeps: srcdg_dep="+srcdg_dep.file_name+", base_cls="+base_cls);

								if(
									srcdg_dep.pkg_name.length() > 0
									&&
									base_cls.startsWith(srcdg_dep.pkg_name)
								  )
									base_cls =
											base_cls
											.substring(srcdg_dep.pkg_name.length()+1);

								tgt = (Tgt)srcdg_dep.targets.Get(base_cls);

								if($.dbug)
									PrintDbugln("ResolveDeps: "+tgt.Dump());

								base_cls = tgt.BaseCls();
							}
						}//end while
					}
				}
			}// end itr for
		}// end s_itr for
	}

	private String FilterClsFromImps(String unpfxd_tgt_name)
	{
		String	pfxd_tgt_name = unpfxd_tgt_name;

		for(
			Enumeration itr=unpfxd_tgt_to_pfxd_tgt_map_.values(unpfxd_tgt_name);
			itr.hasMoreElements();
		   )
		{
			pfxd_tgt_name = (String)itr.nextElement();
			break;
		}

		if($.dbug)
			PrintDbugln("FilterClsFromImps: pfxd_tgt_name="+pfxd_tgt_name);

		return pfxd_tgt_name;
	}

	private SrcDG FindSrcFromCls(String cls_name)
	{
		SrcDG	srcdg = null;
		String	src_name = null;

		if((src_name = (String)tgts_to_src_map_.get(cls_name)) != null)
		{
			srcdg = (SrcDG)srcdg_set_.Get(src_name);
		}

		return srcdg;
	}

	public void ShowGraph()
	{
		SrcDG	srcdg = null;

		for(
			Enumeration itr = srcdg_set_.Iterator();
			itr.hasMoreElements();
			)
		{
			srcdg = (SrcDG)itr.nextElement();

			if($.srcgph)
			{
				PrintMake("\n"+srcdg.file_name+" :");
				ShowSrcDeps(srcdg);
				PrintMake("\n");
			}
			else
			{
				PrintMake("\n");
				if(ShowTargets(srcdg) == true)
				{
					PrintMake(":");
					ShowDeps(srcdg);
					PrintMake("\n");

					if($.rules.length() > 0)
					{
						PrintMake(action_rules_.toString());
						PrintMake("\n");
					}
				}
			}

			if($.nowarn == false && srcdg.warnings.Length() > 0)
			{
				PrintMake("\n"+srcdg.warnings.toString());
			}
		}
	}

	private boolean ShowTargets(SrcDG srcdg)
	{
		boolean rc = false;

		Tgt target = null;
		String	target_name = null;
		String	target_pth = null;
		boolean	is_inner;

		for(
			Enumeration itr = srcdg.targets.Iterator();
			itr.hasMoreElements();
			)
		{
			target = (Tgt)itr.nextElement();
			target_name = target.Name();

			is_inner = (target_name.indexOf('$') < 0 ? false : true);

			if($.noinner == false && is_inner == true)
				target_name = EscInner(target_name);

			if($.noinner == false || is_inner == false)
			{
				target_pth = srcdg.pkg_pth+target_name+".class ";

				PrintMake(target_pth);
				rc = true;

				all_rule_.Append(target_pth);
			}
		}

		if(rc == false)
			WarningsAppend(srcdg,
				"["+srcdg.file_name+"] skipping rule, no targets found"
				);

		return rc;
	}

	private String EscInner(String cls_name)
	{
		String ret_cls_name = null;

		if($.noescinner == true)
		{
			ret_cls_name = cls_name;
		}
		else
		{
			int	di = -1, ni;

			StrBuf tmp_cls_name = new StrBuf();

			while((di = cls_name.indexOf('$', (ni=di+1))) >= 0)
			{
				tmp_cls_name.Append(cls_name.substring(ni, di).concat("$$"));
			}

			ret_cls_name =
				tmp_cls_name.Append(cls_name.substring(ni)).toString();
		}

		return ret_cls_name;
	}

	private void ShowDeps(SrcDG srcdg)
	{
		Dep	dep = null;
		String	dep_name = null;

		for(
			Enumeration itr = srcdg.dep_set.Iterator();
			itr.hasMoreElements();
			)
		{
			dep = (Dep)itr.nextElement();

			dep_name = dep.Name();

			if(use_abspkgmnt_)
			{
				if(dep_name.startsWith(pkg_mnt_))
					dep_name =
						pkg_mnt_canon_+dep_name.substring(pkg_mnt_.length());
			}
			else
			{
				if(dep_name.startsWith(pkg_mnt_canon_))
					dep_name =
						pkg_mnt_+dep_name.substring(pkg_mnt_canon_.length());
			}

			if(dep.State() == Dep.RESOLVED)
			{
				PrintMake(" "+EscInner(dep_name));
			}
			else
			if(dep.State() == Dep.UNRESOLVED)
			{
				String	err_src = dep.GetType().FileName();
				SrcDG err_srcdg = (SrcDG)srcdg_set_.Get(err_src);

				WarningsAppend(err_srcdg,
					"["+dep.GetType().FileAndLine()+"] "
					+"Cannot resolve dependency for possible type "
					+dep_name
					+", so none created"
					);
			}
		}
	}

	private void ShowSrcDeps(SrcDG srcdg)
	{
		Dep		dep = null;
		String	src_name = null;
		HashSet src_set = new HashSet();

		for(
			Enumeration itr = srcdg.dep_set.Iterator();
			itr.hasMoreElements();
			)
		{
			dep = (Dep)itr.nextElement();

			if(dep.State() == Dep.RESOLVED)
			{
				if((src_name = (String)tgts_to_src_map_.get(dep.Name())) != null)
				{
					src_set.add(src_name);
				}
			}
		}

		src_name = null;

		for(
			Enumeration itr = src_set.elements();
			itr.hasMoreElements();
		   )
		{
			src_name = (String)itr.nextElement();
			PrintMake(" "+src_name);
		}
	}

	public void ShowAllRule()
	{
		if(
			$.rules.length() > 0
			&&
			all_rule_.Length() > all_rule_start_len_
		)
			PrintMakeln(all_rule_.toString());
	}

	public void AddPfxdDep(Type type)
	{
		String	poss_cls_name = null;
		String	pkg_name = null;
		String	cls_name = null;
		Class	cls = null;
		String	full_cls_name = null;
		Pkg		pkg = null;
		boolean pkg_in_clspth = false;
		int		oci = 0;

		if($.dbug)
			PrintDbugln("AddPfxdDep: "+type.Dump());

		// remove prefix if its this package

		PruneLocalPkgPfx(type);

		// if we are not checking CLASSPATH then add dep and return

		if($.noclspth)
		{
			AddDep(type);
			return;
		}

		// build a node name stack for type

		ext_nodes_.removeAllElements();

		StringTokenizer type_toker =
			new StringTokenizer(type.Name(), ".");

		StrBuf	tmp_pkg_name = new StrBuf();

		while(type_toker.hasMoreTokens())
		{
			pkg_name = type_toker.nextToken();
			if(tmp_pkg_name.Length() > 0)	tmp_pkg_name.Append(".");
			tmp_pkg_name.Append(pkg_name);
			ext_nodes_.addElement(tmp_pkg_name.toString());
		}

		// reduce nodes by 1 if pri_pfx because last node is method or field

		if(type.IsPriPfx() && ext_nodes_.size() > 1)
			ext_nodes_.removeElementAt(ext_nodes_.size()-1);

		poss_cls_name = (String)ext_nodes_.lastElement();

		// traverse in reverse order and test for class existance

		for(int i=ext_nodes_.size()-1; i >= 0; i--)
		{
			full_cls_name = (String)ext_nodes_.elementAt(i);

			cls = null;

			for(;;)
			{
				try
				{
					cls = Class.forName(FullClsName(full_cls_name));
					break;
				}
				catch(ClassNotFoundException e)
				{
					if(full_cls_name.lastIndexOf('.') < 1)
						break;
				}

				full_cls_name = ReplaceLast(full_cls_name, ".", "$");
			}

			if(cls != null)
			{
				if($.dbug)
					PrintDbugln("AddPfxdDep: forName="+cls.getName());

				if((oci = full_cls_name.lastIndexOf('.')) > 0)
					cls_name = full_cls_name.substring(oci+1);
				else
					cls_name = full_cls_name;

				type.SetSfxdName(cls_name+".class");

				if(oci < 1)	// local or imported class
				{
					AddDep(type);
				}
				else
				{
					pkg_name = (String)ext_nodes_.elementAt(i-1);
					pkg = new Pkg(pkg_name, type.FileName(), type.LineNum());

					if(FindPkgInClsPth(pkg) == true)
					{
						MapClsFromImpToPkg(type.SfxdName(), pkg.LastPkgLoc());
						AddDep(type);
					}
					else
					{
						// major bug because pkg is in CLASSPATH!
					}
				}

				break;
			}
		}

		if(cls_name != null)
		{
			if($.dbug)
				PrintDbugln("cls_name="+cls_name);
		}
		else
		{
			// not found in CLASSPATH, try extpkgs

			String	pkg_node_pth = null;

			for(int i=ext_nodes_.size()-1; i >= 0; i--)
			{
				full_cls_name = (String)ext_nodes_.elementAt(i);

				if((pkg_node_pth = (String)ext_pkgs_.get(full_cls_name)) != null)
				{
					if($.dbug)
						PrintDbugln("pkg_node_pth="+pkg_node_pth);
					break;
				}
			}

			if(pkg_node_pth != null)
			{
				pkg = new Pkg(full_cls_name, type.FileName(), type.LineNum());
				pkg.AddPkgLoc(new File(pkg_node_pth));

				String	full_cls_type =
									type.Name()
									.substring(full_cls_name.length()+1)
									.replace('.','$')
									.concat(".class")
									;

				type.SetSfxdName(full_cls_type);

				MapClsFromImpToPkg(full_cls_type, pkg.LastPkgLoc());
				AddDep(type);
			}
			else
			{
				// not in etxpkgs

				AddDep(type);
			}
		}
	}

	private final String ReplaceLast(String str, String from, String to)
	{
		String  ret_str = str;
		int idx = 0;
		int from_length = from.length();

		if(from_length > 0 && (idx = str.lastIndexOf(from)) > 0)
		{
			ret_str =
				str
				.substring(0, idx)
				.concat(to)
				.concat(str.substring(idx+from_length))
				;
		}

		return ret_str;
	}

	private final void PruneLocalPkgPfx(Type type)
	{
		String pkg_name = srcdg_.pkg_name;

		if(pkg_name.length() > 0)
		{
			String cls_name = type.SfxdName(); 

			if(cls_name.startsWith(pkg_name))
				cls_name = cls_name.substring(pkg_name.length()+1);

			type.SetSfxdName(cls_name);
		}
	}

	public void AddDep(Type type)
		throws
		IllegalStateException,
		IllegalArgumentException
	{
		if(srcdg_ == null)
		{
			throw
			new IllegalStateException("AddSource() must proceed AddDep()");
		}

		if(type == null)
			throw new IllegalArgumentException("missing dependency");

		Dep dep = null;
		String	full_pth = null;

		if(type.Sfx().equals("java"))
		{
			dep = new Dep(type.SfxdName());
		}
		else
		if($.noclspth)
		{
			dep = new Dep(type);
		}
		else
		{
			PkgLoc pkg_loc = FindPkgLoc(type.SfxdName());
			Pkg pkg = (pkg_loc == null ? null : pkg_loc.GetPkg());

			if(pkg != null)
			{
				String pkg_sub_pth = pkg.SubPth(); 
				int si = pkg_sub_pth.length();

				while(
						--si > 1
						&&
						(si = pkg_sub_pth.lastIndexOf(file_sep_, si)) >= 0
					)
				{
					pkg_sub_pth = pkg_sub_pth.substring(0, si+1);

					if(xcld_pkgs_.get(pkg_sub_pth) != null)
						break;
				}

				if(si < 2)	// pkg_sub_pth not in exclude list
				{
					if(pkg_loc.IsZippedDirEnt() && $.nozip == false)
					{
						// the dep is a path within a zip

						dep = new Dep(pkg_loc.ArchPth());
					}
					else
					{
						// the dep is an explicit path

						full_pth = pkg_loc.FullPth();

						if(full_pth.equals(cwd_))
							full_pth = "";

						dep = new Dep(full_pth+type.SfxdName());
					}
				}
			}
			else
			if((full_pth = FindNonPkgClsDir(type.SfxdName())) != null)
			{
				// the dep is a non-packaged class in the CLASSPATH

				dep = new Dep(full_pth+type.SfxdName());
			}
			else
			if(IsSuperVar(type))
			{
				// ignore this dep, it is a var from the super class
			}
			else
			{
				/*
					At this point the dep may reside in an uncompiled
					.java source file. An attempt to resolve it will
					be made later.
				*/

				dep = new Dep(type);
			}
		}

		if(dep != null)
		{
			srcdg_.dep_set.Append(dep);
		}
	}

	public final PkgLoc FindPkgLoc(String cls_name)
	{
		return (PkgLoc)imports_.get(cls_name);
	}

	public final String FullClsName(String cls_name)
	{
		String full_cls_name = cls_name;

		PkgLoc pkg_loc = FindPkgLoc(cls_name+".class");

		if(pkg_loc != null)
		{
			full_cls_name = pkg_loc.GetPkg().Name()+"."+cls_name;
		}

		return full_cls_name;
	}

	private boolean IsSuperVar(Type type)
	{
		boolean rc = false;

		try
		{
			Class cls = Class.forName(type.BaseCls());

			Field fld = null;

			while(cls != null)
			{
				try
				{
					fld = cls.getDeclaredField(type.Name());
					if($.dbug)
						PrintDbugln("getDeclaredField="+fld.getName());
					rc = true;
					break;
				}
				catch(NoSuchFieldException e)
				{
					cls = cls.getSuperclass(); 
					if($.dbug)
						PrintDbugln("getSuperclass="+cls);
				}
			}
		}
		catch(ClassNotFoundException e)
		{
			// do nothing here!
		}
		catch(IllegalArgumentException e)
		{
			// do nothing here!
		}

		if($.dbug)
			PrintDbugln("IsSuperVar: "+rc+" "+type.Dump());

		return rc;
	}

	public boolean FindPkgInClsPth(Pkg pkg)
	{
		boolean rc = false;

		String	pkg_sub_pth = pkg.SubPth();

		for(
			Enumeration itr = class_dir_set_.Iterator();
			itr.hasMoreElements();
		   )
		{
			// fetch next entry in CLASSPATH

			Object	cls_pth_pkg_mnt  = itr.nextElement();

			if(cls_pth_pkg_mnt instanceof HashableZipFile)
			{
				HashableZipFile	hz = (HashableZipFile)cls_pth_pkg_mnt;

				// does the .zip or .jar entry contain the pkg sub pth ?

				if(hz.getEntry(pkg_sub_pth.replace(file_sep_.charAt(0), '/')) != null)
				{
					try
					{
						pkg.AddPkgLoc(new ZippedDir(hz, pkg_sub_pth));

						rc = true;
					}
					catch(IOException t)
					{
						// sub path in hz but not in zd ?
						PrintErrln("# sub path in hz but not in zd");
						PrintErrln(t);
						rc = false;
						break;
					}
				}
			}
			else
			{
				String	pkg_full_pth = (String)cls_pth_pkg_mnt + pkg_sub_pth;
				File	pkg_full_pth_as_file = new File(pkg_full_pth);

				if(pkg_full_pth_as_file.exists())
				{
					pkg.AddPkgLoc(pkg_full_pth_as_file);
					rc = true;
				}
			}
		}

		return rc;
	}

	public void AddPackage(String package_name, String file_name, int lno)
		throws IllegalStateException
	{
		if(srcdg_ == null)
		{
			throw
			new IllegalStateException("AddSource() must proceed AddPackage()");
		}

		if($.prof) PrintProfln("--- Package ---");

		if(package_name != null && package_name.length() > 0)
		{
			if($.prof) PrintProfln("package_name="+package_name);

			srcdg_.pkg_name = package_name;

			String pkg_dir =
						package_name
						.replace('.', file_sep_.charAt(0))
						.concat(file_sep_);	

			srcdg_.pkg_pth = (use_abspkgmnt_ ? pkg_mnt_canon_ : pkg_mnt_);
			
			if(srcdg_.pkg_pth.length() > 0)
				srcdg_.pkg_pth = srcdg_.pkg_pth+pkg_dir;

			try
			{
				AddImport(new Imp(package_name+".*", file_name, lno));
			}
			catch(MissingClsPthEntException e)
			{
				Pkg pkg = e.GetPkg();

				srcdg_.unres_pkg = pkg;
				pkg.AddPkgLoc(new File(srcdg_.pkg_pth));

				WarningsAppend(
						"["+pkg.FileAndLine()+"] import "
						+e.getMessage()
						+" Creating pseudo-package."
					);
			}
		}
	}

	public void AddImport(Imp imp)
		throws
		MissingClsPthEntException
	{
		String import_name = imp.Name();

		int		cls_idx;

		if($.prof)
		{
			if(add_import_called_ == 0)
			{
				add_import_called_ = 1;
				PrintProfln("--- Imports ---");
			}

			if(add_import_called_ > 0)
				PrintProfln("import_name="+import_name);
		}

		if((cls_idx=import_name.lastIndexOf('.')) >= 0)
		{
			String pkg_name =
						import_name
						.substring(0, cls_idx);

			if($.dbug) PrintDbugln("pkg_name="+pkg_name);

			String cls_name =
						import_name
						.substring(++cls_idx)
						.concat(".class");

			if($.dbug) PrintDbugln("cls_name="+cls_name);
			
			Pkg pkg = new Pkg(pkg_name, imp.FileName(), imp.LineNum());

			if(FindPkgInClsPth(pkg) == false)
			{
				throw
				new MissingClsPthEntException(pkg);
			}

			for(
				Enumeration pkgitr = pkg.PkgLocIterator();
				pkgitr.hasMoreElements();
			   )
			{
				PkgLoc	pkg_loc = (PkgLoc)pkgitr.nextElement();

				if(cls_name.startsWith("*"))
				{
					if(pkg_loc.IsFileDirEnt())
					{
						String files[] = pkg_loc.FileDirEnt().list();

						for(int i=0; i < files.length; i++)
						{
							if($.dbug)
								PrintDbugln(
									"files["+i+"]="+files[i]);

							if(files[i].endsWith(".class"))
							{
								MapClsFromImpToPkg(files[i], pkg_loc);
							}
						}
					}
					else
					{
						String	  zcls = null;

						for(
								Enumeration itr = pkg_loc.ZippedDirEnt().Classes();
								itr.hasMoreElements();
						   )
						{
							zcls = (String)itr.nextElement();

							MapClsFromImpToPkg(zcls, pkg_loc);
						}
					}
				}
				else
				{
					MapClsFromImpToPkg(cls_name, pkg_loc);
				}
			}
		}
	}

	public void AddSource(String file_name)
	{
		srcdg_ = new SrcDG(file_name);
		warnings_ = srcdg_.warnings;
		srcdg_set_.Append(srcdg_);
	}

	public void AddTargets(SequencedSet class_decls)
		throws
		IllegalStateException
	{
		if(srcdg_ == null)
		{
			throw
			new IllegalStateException("AddSource() must proceed AddTargets()");
		}

		// add current source file name as a dependency

		AddDep(new Type(srcdg_.file_name, "", "", srcdg_.file_name, 0));

		srcdg_.targets = class_decls.Clone();
		MapTgtsToSrc(srcdg_);
	}

	public void AddVars(HashSet var_set)
		throws
		IllegalStateException
	{
		if(srcdg_ == null)
		{
			throw
			new IllegalStateException("AddSource() must proceed AddTargets()");
		}

		srcdg_.var_set = var_set;
	}

	private final void MapTgtsToSrc(SrcDG srcdg)
	{
		Tgt tgt = null;
		String unpfxd_tgt_name = null;
		String pfxd_tgt_name = null;

		for(
			Enumeration itr = srcdg_.targets.Iterator();
			itr.hasMoreElements();
		   )
		{
			tgt = (Tgt)itr.nextElement();

			pfxd_tgt_name = unpfxd_tgt_name = tgt.Name()+".class";

			if(srcdg_.pkg_name.length() > 0)
				pfxd_tgt_name = srcdg_.pkg_name+"."+unpfxd_tgt_name;

			if($.dbug)
				PrintDbugln("pfxd_tgt_name="+pfxd_tgt_name);

			tgts_to_src_map_.add(pfxd_tgt_name, srcdg_.file_name);

			// note: one key to many values allowed

			if(srcdg_.pkg_name.length() > 0)
				unpfxd_tgt_to_pfxd_tgt_map_
					.add(unpfxd_tgt_name, pfxd_tgt_name);

			if(
				srcdg_.unres_pkg != null
				&&
				FindPkgLoc(unpfxd_tgt_name) == null
			  )
			{
				MapClsFromImpToPkg(unpfxd_tgt_name, srcdg_.unres_pkg.LastPkgLoc());
			}
		}
	}

	private final void MapClsFromImpToPkg(String imp_cls_name, PkgLoc pkg_loc)
	{
		if($.dbug)
			PrintDbugln(
					"imp_cls_name="+imp_cls_name
					+", pkg_full_pth="+pkg_loc.FullPth()
				);

		// map class_name to package on first CLASSPATH occurace only

		imports_.add(imp_cls_name, pkg_loc);
	}

	private final void MapNonPkgClssToDir(String class_dir)
	{
		File	f = new File(class_dir);

		if(f.exists())
		{
			String	clss[] = f.list(cls_sfx_filter_);
			String	cls_dir = (class_dir.equals(cwd_) ? "" : class_dir);

			for(int i=0; i < clss.length; i++)
			{
				non_pkg_cls_dir_map_.add(clss[i], cls_dir);
			}
		}
		else
		{
			if($.nowarn == false)
				PrintErrln(
						"# WARNING: CLASSPATH entry "
						+class_dir.substring(0, class_dir.length()-1)
						+" does not exist."
					);
		}
	}

	private final String FindNonPkgClsDir(String cls_name)
	{
		return (String)non_pkg_cls_dir_map_.get(cls_name);
	}

	public DependencyGraph()
		throws IOException, FileNotFoundException
	{
		if($.dbug)
		{
			$.shwclspth = true;
			$.prof = true;
			$.nowarn = false;

			$.List(DbugWriter());
		}

		srcdg_set_ = new SequencedSet();
		imports_ = new HashMap();
		class_dir_set_ = new SequencedSet();
		ext_nodes_ = new Vector();
		non_pkg_cls_dir_map_ = new HashMap();
		cls_sfx_filter_ = new ClsSfxFilter();
		tgts_to_src_map_ = new HashMap();
		unpfxd_tgt_to_pfxd_tgt_map_ = new HashMap(true);
		warnings_ = new StrBuf();
		all_rule_ =
			new StrBuf(
						"\n"
						+
						($.allrule.length() > 0 ? $.allrule : "all")
						+" : "
						);
		all_rule_start_len_ = all_rule_.Length();

		Reset();

		file_sep_ = System.getProperty("file.separator");
		path_sep_ = System.getProperty("path.separator");
		class_path_ = System.getProperty("java.class.path");
		cwd_ = System.getProperty("user.dir");

		if($.abspkgmnt.length() > 0)
		{
			$.pkgmnt = $.abspkgmnt;
			use_abspkgmnt_ = true;
		}

		if($.pkgmnt.length() > 0)
		{
			pkg_mnt_ =
				($.pkgmnt.endsWith(file_sep_) ? $.pkgmnt : $.pkgmnt+file_sep_);

			try
			{
				File tmp_file = new File(pkg_mnt_);

				if(tmp_file.isDirectory())
				{
					pkg_mnt_canon_ = tmp_file.getCanonicalPath();

					if(pkg_mnt_canon_.endsWith(file_sep_) == false)
						pkg_mnt_canon_ = pkg_mnt_canon_+file_sep_;
						
				}
			}
			catch(NullPointerException e)
			{
				pkg_mnt_canon_ = "";
			}
		}
		else
		{
			pkg_mnt_ = "";
			pkg_mnt_canon_ = "";
		}

		action_rules_ = new StrBuf("");
		pre_rules_ = new StrBuf("");
		post_rules_ = new StrBuf("");

		if(!cwd_.endsWith(file_sep_))
		{
			cwd_ = cwd_ + file_sep_;
		}

		if($.dbug)
			PrintDbugln("cwd_="+cwd_);

		if($.rules.length() > 0)
		{
			if($.rules.equals("default"))
			{
				if($.prerules.length() > 0)
				{
					FileInputStream	fips_prerules =
							new FileInputStream($.prerules);
					pre_rules_.Assign(fips_prerules);	
					PrintMakeln(pre_rules_);
				}
				else
				{
					PrintMake(".SUFFIXES:\n.PRECIOUS: %.java\n\nJCC=javac\n");
				}

				action_rules_.Assign("\t$(JCC)");

				if(pkg_mnt_.length() > 0)
					action_rules_
					.Append(
							" -d "
							+
							(use_abspkgmnt_ ? pkg_mnt_canon_ : pkg_mnt_)
							);

				action_rules_.Append(" $<");
			}
			else
			{
				FileInputStream	fips_rules = new FileInputStream($.rules);
				action_rules_.Assign(fips_rules);	
			}
		}

		if($.shwclspth)
			PrintDbugln("# CLASSPATH="+class_path_);

		xcld_pkgs_ = new HashSet();

		StringTokenizer xcld_pkgs_toker = $.Tokenize("xpkgs");
			
		while(xcld_pkgs_toker.hasMoreTokens())
		{
			String xpkg =	xcld_pkgs_toker
							.nextToken()
							.replace('.', file_sep_.charAt(0))
							.concat(file_sep_)
							;

			xcld_pkgs_.add(xpkg);
		}

		ext_pkgs_ = new HashMap();

		StringTokenizer ext_pkgs_toker = $.Tokenize("extpkgs");
		String	ext_pkg = null,
				ext_pkg_full_pth = null,
				ext_pkg_list = null
				;

		while(ext_pkgs_toker.hasMoreTokens())
		{
			String ext_pkg_tuple = ext_pkgs_toker.nextToken();
			int sepi = ext_pkg_tuple.indexOf(':');

			if(sepi > 0)
			{
				ext_pkg_full_pth =ext_pkg_tuple.substring(0, sepi);

				if(ext_pkg_full_pth.endsWith(file_sep_) == false)
					ext_pkg_full_pth = ext_pkg_full_pth.concat(file_sep_);

				ext_pkg_list = ext_pkg_tuple.substring(sepi+1);

				StringTokenizer ext_pkg_list_toker =
									new StringTokenizer(ext_pkg_list, ":");

				while(ext_pkg_list_toker.hasMoreTokens())
				{
					ext_pkg = ext_pkg_list_toker.nextToken();

					ext_pkg_full_pth =
								ext_pkg_full_pth
								.concat(
										ext_pkg
										.replace('.', file_sep_.charAt(0))
										)
								.concat(file_sep_)
								;

					ext_pkgs_.add(ext_pkg, ext_pkg_full_pth);
				}
			}
			else
			{
				PrintErrln("# -Dextpkgs="+ext_pkg_tuple+" is invalid");
			}
		}
		
		StringTokenizer class_toker =
			new StringTokenizer(class_path_, path_sep_);

		/*
			Iterate through each path in the CLASSPATH and
			add the path to a unique set. If the path is
			a zip file then make it hashable before adding
			it to the set.
		*/

		while(class_toker.hasMoreTokens())
		{
			String class_dir = class_toker.nextToken().trim();

			if(class_dir.length() < 1)
				continue;
			else
			if(
				class_dir.endsWith(".zip")
				||
				class_dir.endsWith(".jar")
				)
			{
				try
				{
					class_dir_set_
					.Append(new HashableZipFile(class_dir));
				}
				catch(IOException e)
				{
					if($.nowarn == false)
						PrintErrln(
									"# WARNING: CLASSPATH entry "
									+class_dir
									+" does not exist."
								);
				}

				continue;
			}

			if(class_dir.equals("."))
			{
				class_dir = cwd_;
			}
			else
			if(class_dir.endsWith(file_sep_) == false)
			{
				class_dir = class_dir+file_sep_;
			}

			class_dir_set_.Append(class_dir);
			MapNonPkgClssToDir(class_dir);
		}

		add_import_called_ = -1;

		try
		{
			AddImport(new Imp("java.lang.*", "", 0));
		}
		catch(MissingClsPthEntException e)
		{
			PrintErrln(e.getMessage());
		}
	}
}
//------------------------------------------
