/*
    =========================================================================
	File:		GetProps.java
	Author:		Warren Pfeffer
	Contact:	slinky@njcc.com
    =========================================================================
    Copyright (C) 1997  Warren Pfeffer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program (see COPYING-2.0.gz); if not, write to:

		Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307  USA
    =========================================================================
*/
package WARREN.tools;

/*
	Here are the subclassing guidlines:

	o	The subclass must be public because
		it will not reside within this pacakge!
		In order for GetProps to modify a subclasses
		fields it must be a public subclass. The
		instance of this subclass can be private
		if you wish.

	o	The subclass should only contain
		public fields of type boolean or String.

	o	By default, boolean fields are set to
		false, and String fields are set to an
		empty string "".

	o	A property flag can be scoped to the declaring
		class of the subclass of GetProps by prepending
		the declaring class name (followed by an underscore)
		to the public field in the subclass being targeted.
		See GetPropsTest.java for an example.
*/

import java.lang.reflect.*;
import java.util.Properties;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.StringTokenizer;

public class GetProps {

	private	Class	cls_;
	private boolean	dbug_;
	private Properties props_;
	private String	class_heir_name_;	// full class name
	private String  decl_class_name_;
	private Field	flds_[];

	protected GetProps()
	{
		this(System.getProperties(), false);
	}

	protected GetProps(boolean deebug)
	{
		this(System.getProperties(), deebug);
	}

	protected GetProps(Properties props, boolean deebug)
	{
		int	idx; // tmp index

		props_ = props; 
		dbug_ = deebug;

		cls_ = getClass();

		if(dbug_) System.err.println("cls_="+cls_);

		class_heir_name_ = cls_.getName();

		idx = class_heir_name_.lastIndexOf('.') + 1;

		class_heir_name_ = class_heir_name_.substring(idx);

		if(dbug_) System.err.println("class_heir_name_="+class_heir_name_);

		if((idx = class_heir_name_.lastIndexOf('$')) >= 0)
        {
			decl_class_name_ = class_heir_name_.substring(0, idx);
			
			if((idx = decl_class_name_.lastIndexOf('$')) >= 0)
			{
				decl_class_name_ = decl_class_name_.substring(idx+1);
			}
        }
		else
		{
			decl_class_name_ = class_heir_name_;
		}

		if(dbug_) System.err.println("decl_class_name_="+decl_class_name_);

		flds_ = cls_.getFields();

		if(System.getProperty("help") != null)
		{
			ShowHelp();
			System.exit(0);
		}
		else
		{
			SetProps();
		}
	}

	private void SetProps()
	{
		String	prop = null;
		Field	fld = null;
		String	fld_name = null;

		for(int i=0; i < flds_.length; i++)
		{
			fld = flds_[i];

			if(Modifier.isPublic(fld.getModifiers()))
			{
				fld_name = fld.getName();

				if(dbug_) System.err.println("fld_name="+fld_name);

				if((prop = props_.getProperty(decl_class_name_+"_"+fld_name)) == null)
				{
					prop = props_.getProperty(fld_name);
				}

				if(dbug_ && prop == null) System.err.println("prop=null");

				try
				{
					if(fld.getType().getName().equals("boolean"))
					{
						fld.setBoolean(this, (prop == null ? false : true));
					}
					else
					if(fld.getType().getName().equals("java.lang.String"))
					{
						fld.set(this, (prop == null ? "" : prop.trim()));
					}
					else
					{
						// we should throw an exception here
					}
				}
				catch(IllegalAccessException e)
				{
					System.err.println("field="+fld_name+" "+e);
				}
			}
		}
	}

	private void ShowHelp()
	{
		String	help_text = null;
		Field	fld = null;

		for(int i=0; i < flds_.length; i++)
		{
			fld = flds_[i];

			if(Modifier.isPublic(fld.getModifiers()))
			{
				help_text = "-D"+decl_class_name_+"_"+fld.getName(); 

				if(fld.getType().getName().equals("boolean") == false)
				{
					help_text = help_text+"=";
				}

				System.err.println(help_text);
			}
		}
	}

	public void List(PrintStream ps)
	{
		List(new PrintWriter(ps));
	}

	public void List(PrintWriter pw)
	{
		String	val = null;
		Field	fld = null;

		for(int i=0; i < flds_.length; i++)
		{
			fld = flds_[i];

			if(Modifier.isPublic(fld.getModifiers()))
			{
				try
				{
					val = fld.get(this).toString();
				}
				catch(IllegalAccessException e)
				{
					val = "?";
				}

				pw.println(
					class_heir_name_
					+"."
					+fld.getName()
					+"="
					+val);
			}
		}
	}

	public StringTokenizer Tokenize(String fld_name)
	{
		StringTokenizer st = null;
		Field			fld = null;

		try
		{
			if(
				(fld = cls_.getField(fld_name)) != null
				&& 
				fld.getType().getName().equals("java.lang.String")
			  )
			{
				st = new StringTokenizer((String)fld.get(this), ", \t");
			}
		}
		catch(Exception e)
		{
			st = null;
		}

		return st;
	}
}
